/* Soot - a J*va Optimization Framework
 * Copyright (C) 2004 Jennifer Lhotak
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

package soot.javaToJimple.ppa.jj;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import polyglot.ast.NodeFactory;
import polyglot.frontend.Job;
import polyglot.frontend.Pass;
import polyglot.types.LoadedClassResolver;
import polyglot.types.SemanticException;
import polyglot.types.TypeSystem;
import polyglot.util.InternalCompilerError;
import soot.javaToJimple.ppa.jj.ast.PPANodeFactory_c;
import soot.javaToJimple.ppa.jj.types.PPASourceClassResolver;
import soot.javaToJimple.ppa.jj.types.PPATypeSystem;
import soot.options.Options;

/**
 * Extension information for jj extension.
 */
public class PPAExtensionInfo extends soot.javaToJimple.jj.ExtensionInfo {
    static {
        // force Topics to load
        new Topics();
    }
    
    private PPATypeSystem ppaTypeSystem;
    
    public PPAExtensionInfo(PPATypeSystem ppaTypeSystem) {
		super();
		this.ppaTypeSystem = ppaTypeSystem;
	}

	@Override
	public String compilerName() {
        return "ppa.jjc";
    }

    @Override
	protected NodeFactory createNodeFactory() {
        return new PPANodeFactory_c();
    }

    /*public Parser parser(Reader reader, FileSource source, ErrorQueue eq) {
        Lexer lexer = new Lexer_c(reader, source.name(), eq);
        Grm grm = new Grm(lexer, ts, nf, eq);
        return new CupParser(grm, source, eq);
    }*/

    /**
     * Appends the soot classpath to the default system classpath.
     */
    @Override
	protected polyglot.main.Options createOptions() {
    	return new polyglot.main.Options(this) {

			/**
			 * Appends the soot classpath to the default system classpath.
			 */
			@Override
			public String constructFullClasspath() {
				String cp = super.constructFullClasspath();
				cp += File.pathSeparator + soot.options.Options.v().soot_classpath();
				return cp;
			}

    	};
    }

    @Override
	protected TypeSystem createTypeSystem() {
        return ppaTypeSystem;
    }

    @Override
	public String defaultFileExtension() {
        return "ppa.jj";
    }

    @Override
	protected void initTypeSystem() {
		try {
			 PPATypeSystem jTs = (PPATypeSystem) ts;
			if (!jTs.isInitialized()) {
				LoadedClassResolver lr;
				lr = new PPASourceClassResolver(compiler, this, Options.v().soot_classpath(), compiler
						.loader(), true, ts);
				jTs.initialize(lr, this);
			} else {
				jTs.reInitialize(this);
			}
		} catch ( SemanticException e) {
			throw new InternalCompilerError("Unable to initialize type system: " + e.getMessage());
		}
	}

    @Override
	public List passes(Job job) {
         List passes = super.passes(job);
         List<Pass> toRemove = new ArrayList<Pass>();
         for (Object passObject : passes) {
        	 Pass pass = (Pass) passObject;
        	 if (pass.id() == Pass.REACH_CHECK || pass.id() == Pass.CONSTRUCTOR_CHECK || pass.id() == Pass.INIT_CHECK || pass.id() == Pass.FWD_REF_CHECK) {
        		 toRemove.add(pass);
        	 }
         }
         
         passes.removeAll(toRemove);
         
        return passes;
    }

    public Job getCurrentJob() {
    	return currentJob;
    }
}
