package soot.javaToJimple.ppa.jj.types;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import polyglot.ext.jl.types.ClassType_c;
import polyglot.ext.jl.types.NullType_c;
import polyglot.ext.jl.types.Package_c;
import polyglot.ext.jl.types.TypeObject_c;
import polyglot.types.ArrayType;
import polyglot.types.ClassType;
import polyglot.types.ConstructorInstance;
import polyglot.types.FieldInstance;
import polyglot.types.Flags;
import polyglot.types.MethodInstance;
import polyglot.types.NullType;
import polyglot.types.Package;
import polyglot.types.PrimitiveType;
import polyglot.types.ReferenceType;
import polyglot.types.Resolver;
import polyglot.types.Type;
import polyglot.types.TypeObject;
import polyglot.types.TypeSystem;
import polyglot.util.Position;
import soot.javaToJimple.ppa.TypeFactUtil;

public class MagicClass extends TypeObject_c implements ClassType {

	public  static String MAGIC_PACKAGE = "MMAGICPPACKAGE";

	public  static String DEFAULT_MAGIC_CLASS = MAGIC_PACKAGE + ".MagicClass";

	private ReferenceType container;

	private MagicClassArray array;

	private boolean isArray;

	private final  String name;

	private final  String packageName;

	private final  String fullName;

	private final  TypeSystem ts;

	private Flags flags;

	private final  List<ConstructorInstance> constructors;

	private final  List<FieldInstance> fields;

	private final  List<MethodInstance> methods;

	private final  List<ClassType> memberClasses;

	public List<Type> relatedTypes;

	public MagicClass(String name, String packageName, String fullName, TypeSystem ts) {
		super(ts);
		this.name = name;
		this.packageName = packageName;
		this.fullName = fullName;
		this.ts = ts;
		flags = Flags.PUBLIC;
		constructors = new ArrayList<ConstructorInstance>();
		methods = new ArrayList<MethodInstance>();
		fields = new ArrayList<FieldInstance>();
		relatedTypes = new ArrayList<Type>();
		memberClasses = new ArrayList<ClassType>();
	}

	public void addRelatedTypes(Type type) {
		relatedTypes.add(type);
	}

	public ArrayType arrayOf() {
		return null;
	}

	public ArrayType arrayOf(int dims) {
		return null;
	}

	public List constructors() {
		return constructors;
	}

	public ReferenceType container() {
		return container;
	}

	@Override
	public Object copy() {
		return this;
	}

	public boolean descendsFrom(Type ancestor) {
		return true;
	}

	public boolean descendsFromImpl(Type t) {
		return true;
	}

	@Override
	public boolean equalsImpl(TypeObject t) {
		// Need to correctly evaluate equality
		if (t instanceof MagicClass) {
			 MagicClass m = (MagicClass) t;
			return m.fullName().equals(this.fullName());
		}

		return false;
	}

	public FieldInstance fieldNamed(String name) {

		FieldInstance fi = null;
		for ( FieldInstance temp : fields) {
			if (temp.name().equals(name)) {
				fi = temp;
				break;
			}
		}

		return fi;
	}

	public List fields() {
		return fields;
	}

	public Flags flags() {
		return flags;
	}

	public String fullName() {
		return fullName;
	}

	public List<ConstructorInstance> getConstructors() {
		return constructors;
	}

	public List<FieldInstance> getFields() {
		return fields;
	}

	public List<MethodInstance> getMethods() {
		return methods;
	}

	public String getPackageName() {
		return packageName;
	}

	public List<Type> getRelatedTypes() {
		return Collections.unmodifiableList(relatedTypes);
	}

	public boolean hasEnclosingInstance(ClassType encl) {
		return false;
	}

	public boolean hasEnclosingInstanceImpl(ClassType encl) {
		return false;
	}

	public boolean hasMethod(MethodInstance mi) {
		return true;
	}

	public boolean hasMethodImpl(MethodInstance mi) {
		return true;
	}

	public boolean inStaticContext() {
		return false;
	}

	public List interfaces() {
		return new ArrayList();
	}

	public boolean isAnonymous() {
		return false;
	}

	public boolean isArray() {
		return isArray;
	}

	public boolean isBoolean() {
		return false;
	}

	public boolean isByte() {
		return false;
	}

	public boolean isCanonical() {
		return true;
	}

	public boolean isCastValid(Type toType) {
		return true;
	}

	public boolean isCastValidImpl(Type t) {
		return true;
	}

	public boolean isChar() {
		return false;
	}

	public boolean isClass() {
		return true;
	}

	public boolean isComparable(Type t) {
		return false;
	}

	public boolean isDouble() {
		return false;
	}

	public boolean isEnclosed(ClassType outer) {
		return false;
	}

	public boolean isEnclosedImpl(ClassType outer) {
		return false;
	}

	public boolean isFloat() {
		return false;
	}

	public boolean isImplicitCastValid(Type toType) {
		boolean isFinal = false;
		if (toType instanceof ClassType_c) {
			 ClassType_c cType = (ClassType_c) toType;
			isFinal = cType.flags().isFinal();
		}

		// You can always cast a magic class to a default magic class or to any other type that is
		// not a primitive, an array or a  type...

		// XXX Maybe we should add the type info to the related type???

		return fullName.equals(DEFAULT_MAGIC_CLASS)
				|| (!toType.isPrimitive() && !toType.isArray() && !isFinal);
		// return true;
	}

	public boolean isImplicitCastValidImpl(Type t) {
		// for now... same...
		// FIXME This should be different (the opposite? of isImplicitCastValid)
		return isImplicitCastValid(t);
	}

	public boolean isInner() {
		return false;
	}

	public boolean isInnerClass() {
		return false;
	}

	public boolean isInt() {
		return false;
	}

	public boolean isIntOrLess() {
		return false;
	}

	public boolean isLocal() {
		return false;
	}

	public boolean isLong() {
		return false;
	}

	public boolean isLongOrLess() {
		return false;
	}

	public boolean isMagicCastValid(Type type) {
		boolean castValid = false;

		if (TypeFactUtil.isAnonymousMagicClass(this)) {
			castValid = true;
		} else {
			if (type instanceof MagicClass && !relatedTypes.contains(type)) {
				relatedTypes.add(type);
			}

			// FIXME: WHAT ABOUT TYPE THAT HAVE MAGIC HIERARCHY???
			for ( Type relatedType : relatedTypes) {
				if (type.equalsImpl(relatedType)) {
					castValid = true;
					break;
				}
			}
		}

		return castValid;
	}

	public boolean isMember() {
		return false;
	}

	public boolean isNested() {
		return false;
	}

	public boolean isNull() {
		return false;
	}

	public boolean isNumeric() {
		return false;
	}

	public boolean isPackage() {
		return false;
	}

	public boolean isPrimitive() {
		return false;
	}

	public boolean isReference() {
		return true;
	}

	public boolean isShort() {
		return false;
	}

	public boolean isSubtype(Type ancestor) {
		return true;
	}

	public boolean isSubtypeImpl(Type t) {
		return true;
	}

	public boolean isThrowable() {
		return true;
	}

	public boolean isTopLevel() {
		return true;
	}

	public boolean isType() {
		return false;
	}

	public boolean isUncheckedException() {
		return true;
	}

	public boolean isVoid() {
		return false;
	}

	public Kind kind() {
		return ClassType.TOP_LEVEL;
	}

	public List memberClasses() {
		return memberClasses;
	}

	public ClassType memberClassNamed(String name) {
		ClassType t = null;
		for (Iterator i = memberClasses().iterator(); i.hasNext();) {
			t = (ClassType) i.next();
			if (t.name().equals(name)) {
				break;
			}
		}

		if (t != null && t.name().equals(name)) {
			t = null;
		}

		return t;
	}

	public List methods() {
		return methods;
	}

	public List methods(String name, List argTypes) {
		 List<MethodInstance> list = new ArrayList<MethodInstance>();

		for ( MethodInstance temp : methods) {
			if (temp.name().equals(name) && temp.formalTypes().equals(argTypes)) {
				list.add(temp);
			}
		}

		return list;
	}

	public List methodsNamed(String name) {
		 List<MethodInstance> list = new ArrayList<MethodInstance>();

		for ( MethodInstance temp : methods) {
			if (temp.name().equals(name)) {
				list.add(temp);
			}
		}

		if (list.isEmpty()) {
			// When this method is called, it means that a method of this name MUST exist.
			// On the other hand, it is questionnable whether the method should be added to the
			// list...
			try {
				 MethodInstance mi = new MagicMethodInstance_c(ts, Position.compilerGenerated(),
						this, Flags.PUBLIC, ts.typeForName(DEFAULT_MAGIC_CLASS), name,
						new ArrayList(), new ArrayList());
				list.add(mi);
				// methods.add(mi);
			} catch ( Exception e) {
				// Should not happen
				e.printStackTrace();
			}
		}

		return list;
	}

	public String name() {
		return name;
	}

	public boolean numericConversionValid(long value) {
		return false;
	}

	public boolean numericConversionValid(Object value) {
		return false;
	}

	public boolean numericConversionValidImpl(long value) {
		return false;
	}

	public boolean numericConversionValidImpl(Object value) {
		return false;
	}

	public ClassType outer() {
		return null;
	}

	public Package package_() {
		return new Package_c(ts, packageName);
	}

	@Override
	public Position position() {
		return null;
	}

	public void replaceField(FieldInstance oldFi, FieldInstance newFi) {
		FieldInstance toRemove = null;

		for ( FieldInstance temp : fields) {
			if (temp.name().equals(oldFi.name())) {
				if (!temp.equalsImpl(newFi)) {
					toRemove = temp;
				}
				break;
			}
		}

		if (toRemove != null) {
			fields.remove(toRemove);
			fields.add(newFi);
		}

	}

	public void setContainer(ReferenceType container) {
		this.container = container;
	}

	public void setFlags(Flags flags) {
		this.flags = flags;
	}

	public void setIsArray(boolean isArray) {
		this.isArray = isArray;
	}

	public Type superType() {
		Type type = null;

		try {
			// For now, the super type is always Object.
			// It is often an undecidable problem to try to infer the whole hierarchy.
			type = ts.typeForName("java.lang.Object");
		} catch ( Exception e) {
			e.printStackTrace();
		}

		return type;
	}

	public ArrayType toArray() {
		// XXX This will make all unknown anonymous type an array...
		ArrayType aType = array;

		if (isArray && array == null) {
			array = new MagicClassArray(ts, position, this);
			aType = array;
		}
		return aType;
	}

	public ClassType toClass() {
		return this;
	}

	public NullType toNull() {
		return new NullType_c(ts);
	}

	public Package toPackage() {
		return null;
	}

	public PrimitiveType toPrimitive() {
		return null;
	}

	public ReferenceType toReference() {
		return this;
	}

	@Override
	public String toString() {
		return fullName;
	}

	public Type toType() {
		return this;
	}

	public String translate(Resolver c) {
		return null;
	}

	@Override
	public TypeSystem typeSystem() {
		return ts;
	}

}
